@@ -14,11 +14,12 @@
 // under GPL version 2 or later
 //
 // Copyright (C) 2005 Kristian Høgsberg <krh@redhat.com>
-// Copyright (C) 2005, 2007 Albert Astals Cid <aacid@kde.org>
+// Copyright (C) 2005, 2007, 2009-2011 Albert Astals Cid <aacid@kde.org>
 // Copyright (C) 2005 Jonathan Blandford <jrb@redhat.com>
 // Copyright (C) 2005, 2006, 2008 Brad Hards <bradh@frogmouth.net>
 // Copyright (C) 2007 Julien Rebetez <julienr@svn.gnome.org>
-// Copyright (C) 2008 Pino Toscano <pino@kde.org>
+// Copyright (C) 2008, 2011 Pino Toscano <pino@kde.org>
+// Copyright (C) 2010 Hib Eris <hib@hiberis.nl>
 //
 // To see a description of the changes please see the Changelog file that
 // came with your tarball or type make ChangeLog if you are building from git
@@ -32,6 +33,11 @@
 #pragma interface
 #endif
 
+#include "Object.h"
+
+#include <vector>
+
+class PDFDoc;
 class XRef;
 class Object;
 class Page;
@@ -41,6 +47,8 @@ class LinkDest;
 class PageLabelInfo;
 class Form;
 class OCGs;
+class ViewerPreferences;
+class FileSpec;
 
 //------------------------------------------------------------------------
 // NameTree
@@ -49,12 +57,11 @@ class OCGs;
 class NameTree {
 public:
   NameTree();
+  ~NameTree();
   void init(XRef *xref, Object *tree);
-  void parse(Object *tree);
   GBool lookup(GooString *name, Object *obj);
-  void free();
   int numEntries() { return length; };
-  // iterator accessor
+  // iterator accessor, note it returns a shallow copy, do not free the object
   Object getValue(int i);
   GooString *getName(int i);
 
@@ -68,6 +75,7 @@ private:
     static int cmp(const void *key, const void *entry);
   };
 
+  void parse(Object *tree);
   void addEntry(Entry *entry);
 
   XRef *xref;
@@ -78,57 +86,6 @@ private:
                     // length is the number of real Entry
 };
 
-class EmbFile {
-public:
-  EmbFile(GooString *name, GooString *description, 
-	  int size,
-	  GooString *createDate,
-	  GooString *modDate, GooString *checksum,
-	  GooString *mimetype,
-	  Object objStr) :
-    m_name(name),
-    m_description(description),
-    m_size(size),
-    m_createDate(createDate),
-    m_modDate(modDate),
-    m_checksum(checksum),
-    m_mimetype(mimetype)
-  {
-    objStr.copy(&m_objStr);
-  }
-  EmbFile(Object *efDict, GooString *description = 0);
-
-  ~EmbFile()
-  {
-    delete m_name;
-    delete m_description;
-    delete m_modDate;
-    delete m_createDate;
-    delete m_checksum;
-    delete m_mimetype;
-    m_objStr.free();
-  }
-
-  GooString *name() { return m_name; }
-  GooString *description() { return m_description; }
-  int size() { return m_size; }
-  GooString *modDate() { return m_modDate; }
-  GooString *createDate() { return m_createDate; }
-  GooString *checksum() { return m_checksum; }
-  GooString *mimeType() { return m_mimetype; }
-  Object &streamObject() { return m_objStr; }
-
-private:
-  GooString *m_name;
-  GooString *m_description;
-  int m_size;
-  GooString *m_createDate;
-  GooString *m_modDate;
-  GooString *m_checksum;
-  GooString *m_mimetype;
-  Object m_objStr;
-};
-
 //------------------------------------------------------------------------
 // Catalog
 //------------------------------------------------------------------------
@@ -137,7 +94,7 @@ class Catalog {
 public:
 
   // Constructor.
-  Catalog(XRef *xrefA);
+  Catalog(PDFDoc *docA);
 
   // Destructor.
   ~Catalog();
@@ -146,13 +103,13 @@ public:
   GBool isOk() { return ok; }
 
   // Get number of pages.
-  int getNumPages() { return numPages; }
+  int getNumPages();
 
   // Get a page.
-  Page *getPage(int i) { return pages[i-1]; }
+  Page *getPage(int i);
 
   // Get the reference for a page object.
-  Ref *getPageRef(int i) { return &pageRefs[i-1]; }
+  Ref *getPageRef(int i);
 
   // Return base URI, or NULL if none.
   GooString *getBaseURI() { return baseURI; }
@@ -162,7 +119,7 @@ public:
   GooString *readMetadata();
 
   // Return the structure tree root object.
-  Object *getStructTreeRoot() { return &structTreeRoot; }
+  Object *getStructTreeRoot();
 
   // Find a page, given its object ID.  Returns page number, or 0 if
   // not found.
@@ -172,16 +129,16 @@ public:
   // NULL if <name> is not a destination.
   LinkDest *findDest(GooString *name);
 
-  Object *getDests() { return &dests; }
+  Object *getDests();
 
   // Get the number of embedded files
-  int numEmbeddedFiles() { return embeddedFileNameTree.numEntries(); }
+  int numEmbeddedFiles() { return getEmbeddedFileNameTree()->numEntries(); }
 
   // Get the i'th file embedded (at the Document level) in the document
-  EmbFile *embeddedFile(int i);
+  FileSpec *embeddedFile(int i);
 
   // Get the number of javascript scripts
-  int numJS() { return jsNameTree.numEntries(); }
+  int numJS() { return getJSNameTree()->numEntries(); }
 
   // Get the i'th JavaScript script (at the Document level) in the document
   GooString *getJS(int i);
@@ -190,13 +147,15 @@ public:
   GBool labelToIndex(GooString *label, int *index);
   GBool indexToLabel(int index, GooString *label);
 
-  Object *getOutline() { return &outline; }
+  Object *getOutline();
 
   Object *getAcroForm() { return &acroForm; }
 
   OCGs *getOptContentConfig() { return optContent; }
 
-  Form* getForm() { return form; }
+  Form* getForm();
+
+  ViewerPreferences *getViewerPreferences();
 
   enum PageMode {
     pageModeNone,
@@ -204,7 +163,8 @@ public:
     pageModeThumbs,
     pageModeFullScreen,
     pageModeOC,
-    pageModeAttach
+    pageModeAttach,
+    pageModeNull
   };
   enum PageLayout {
     pageLayoutNone,
@@ -213,39 +173,57 @@ public:
     pageLayoutTwoColumnLeft,
     pageLayoutTwoColumnRight,
     pageLayoutTwoPageLeft,
-    pageLayoutTwoPageRight
+    pageLayoutTwoPageRight,
+    pageLayoutNull
   };
 
   // Returns the page mode.
-  PageMode getPageMode() { return pageMode; }
-  PageLayout getPageLayout() { return pageLayout; }
+  PageMode getPageMode();
+  PageLayout getPageLayout();
 
 private:
 
+  // Get page label info.
+  PageLabelInfo *getPageLabelInfo();
+
+  PDFDoc *doc;
   XRef *xref;			// the xref table for this PDF file
   Page **pages;			// array of pages
   Ref *pageRefs;		// object ID for each page
+  int lastCachedPage;
+  std::vector<Dict *> *pagesList;
+  std::vector<Ref> *pagesRefList;
+  std::vector<PageAttrs *> *attrsList;
+  std::vector<int> *kidsIdxList;
   Form *form;
+  ViewerPreferences *viewerPrefs;
   int numPages;			// number of pages
   int pagesSize;		// size of pages array
   Object dests;			// named destination dictionary
-  NameTree destNameTree;	// named destination name-tree
-  NameTree embeddedFileNameTree;  // embedded file name-tree
-  NameTree jsNameTree;		// Java Script name-tree
+  Object names;			// named names dictionary
+  NameTree *destNameTree;	// named destination name-tree
+  NameTree *embeddedFileNameTree;  // embedded file name-tree
+  NameTree *jsNameTree;		// Java Script name-tree
   GooString *baseURI;		// base URI for URI-type links
   Object metadata;		// metadata stream
   Object structTreeRoot;	// structure tree root dictionary
   Object outline;		// outline dictionary
   Object acroForm;		// AcroForm dictionary
+  Object viewerPreferences;     // ViewerPreference dictionary
   OCGs *optContent;		// Optional Content groups
   GBool ok;			// true if catalog is valid
   PageLabelInfo *pageLabelInfo; // info about page labels
   PageMode pageMode;		// page mode
   PageLayout pageLayout;	// page layout
 
-  int readPageTree(Dict *pages, PageAttrs *attrs, int start,
-		   char *alreadyRead);
+  GBool cachePageTree(int page); // Cache first <page> pages.
   Object *findDestInTree(Object *tree, GooString *name, Object *obj);
+
+  Object *getNames();
+  NameTree *getDestNameTree();
+  NameTree *getEmbeddedFileNameTree();
+  NameTree *getJSNameTree();
+
 };
 
 #endif
