@@ -6,13 +6,14 @@
 //
 // Copyright 2005 Jeff Muizelaar <jeff@infidigm.net>
 // Copyright 2005 Martin Kretzschmar <martink@gnome.org>
-// Copyright 2005-2007 Albert Astals Cid <aacid@kde.org>
+// Copyright 2005-2007, 2009-2011 Albert Astals Cid <aacid@kde.org>
+// Copyright 2010 Carlos Garcia Campos <carlosgc@gnome.org>
+// Copyright 2011 Daiki Ueno <ueno@unixuser.org>
 //
 //========================================================================
 
 #ifndef DCTSTREAM_H
 #define DCTSTREAM_H
-#include <config.h>
 
 #ifdef USE_GCC_PRAGMAS
 #pragma interface
@@ -23,9 +24,11 @@
 #pragma implementation
 #endif
 
+#include "poppler-config.h"
 #include <stdio.h>
 #include <stdlib.h>
 #include <stddef.h>
+#include <setjmp.h>
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
@@ -33,7 +36,6 @@
 #include <ctype.h>
 #include "goo/gmem.h"
 #include "goo/gfile.h"
-#include "poppler-config.h"
 #include "Error.h"
 #include "Object.h"
 #include "Decrypt.h"
@@ -48,9 +50,12 @@ struct str_src_mgr {
     JOCTET buffer;
     Stream *str;
     int index;
-    bool abort;
 };
 
+struct str_error_mgr {
+  struct jpeg_error_mgr pub;
+  jmp_buf setjmp_buffer;
+};
 
 class DCTStream: public FilterStream {
 public:
@@ -61,16 +66,20 @@ public:
   virtual void reset();
   virtual int getChar();
   virtual int lookChar();
-  virtual GooString *getPSFilter(int psLevel, char *indent);
+  virtual GooString *getPSFilter(int psLevel, const char *indent);
   virtual GBool isBinary(GBool last = gTrue);
-  Stream *getRawStream() { return str; }
 
 private:
   void init();
 
-  unsigned int x;
+  virtual GBool hasGetChars() { return true; }
+  virtual int getChars(int nChars, Guchar *buffer);
+
+  int colorXform;
+  JSAMPLE *current;
+  JSAMPLE *limit;
   struct jpeg_decompress_struct cinfo;
-  struct jpeg_error_mgr jerr;
+  struct str_error_mgr err;
   struct str_src_mgr src;
   JSAMPARRAY row_buffer;
 };
