@@ -14,6 +14,7 @@
 // under GPL version 2 or later
 //
 // Copyright (C) 2008 Julien Rebetez <julien@fhtagn.net>
+// Copyright (C) 2009 David Benjamin <davidben@mit.edu>
 //
 // To see a description of the changes please see the Changelog file that
 // came with your tarball or type make ChangeLog if you are building from git
@@ -38,7 +39,6 @@
 
 class Decrypt {
 public:
-  static void md5(Guchar *msg, int msgLen, Guchar *digest);
 
   // Generate a file key.  The <fileKey> buffer must have space for at
   // least 16 bytes.  Checks <ownerPassword> and then <userPassword>
@@ -47,6 +47,7 @@ public:
   // may be NULL, which is treated as an empty string.
   static GBool makeFileKey(int encVersion, int encRevision, int keyLength,
 			   GooString *ownerKey, GooString *userKey,
+			   GooString *ownerEnc, GooString *userEnc,
 			   int permissions, GooString *fileID,
 			   GooString *ownerPassword, GooString *userPassword,
 			   Guchar *fileKey, GBool encryptMetadata,
@@ -79,6 +80,14 @@ struct DecryptAESState {
   int bufIdx;
 };
 
+struct DecryptAES256State {
+  Guint w[60];
+  Guchar state[16];
+  Guchar cbc[16];
+  Guchar buf[16];
+  int bufIdx;
+};
+
 class DecryptStream: public FilterStream {
 public:
 
@@ -90,6 +99,7 @@ public:
   virtual void reset();
   virtual int getChar();
   virtual int lookChar();
+  virtual int getPos();
   virtual GBool isBinary(GBool last);
   virtual Stream *getUndecodedStream() { return this; }
 
@@ -97,12 +107,20 @@ private:
 
   CryptAlgorithm algo;
   int objKeyLength;
-  Guchar objKey[16 + 9];
+  Guchar objKey[32];
+  int charactersRead; // so that getPos() can be correct
 
   union {
     DecryptRC4State rc4;
     DecryptAESState aes;
+    DecryptAES256State aes256;
   } state;
 };
+ 
+//------------------------------------------------------------------------
+
+extern void rc4InitKey(Guchar *key, int keyLen, Guchar *state);
+extern Guchar rc4DecryptByte(Guchar *state, Guchar *x, Guchar *y, Guchar c);
+extern void md5(Guchar *msg, int msgLen, Guchar *digest);
 
 #endif
