@@ -5,8 +5,10 @@
 // This file is licensed under the GPLv2 or later
 //
 // Copyright 2006 Julien Rebetez <julienr@svn.gnome.org>
-// Copyright 2007, 2008 Carlos Garcia Campos <carlosgc@gnome.org>
-// Copyright 2007-2009 Albert Astals Cid <aacid@kde.org>
+// Copyright 2007, 2008, 2011 Carlos Garcia Campos <carlosgc@gnome.org>
+// Copyright 2007-2010 Albert Astals Cid <aacid@kde.org>
+// Copyright 2010 Mark Riedesel <mark@klowner.com>
+// Copyright 2011 Pino Toscano <pino@kde.org>
 //
 //========================================================================
 
@@ -18,13 +20,18 @@
 #endif
 
 #include "Object.h"
-#include "goo/GooVector.h"
+
+#include <set>
 
 class GooString;
 class Array;
 class Dict;
 class Annot;
-class Catalog;
+class AnnotWidget;
+class Annots;
+class LinkAction;
+class GfxResources;
+class PDFDoc;
 
 enum FormFieldType {
   formButton,
@@ -40,6 +47,12 @@ enum FormButtonType {
   formButtonRadio
 };
 
+enum VariableTextQuadding {
+  quaddingLeftJustified,
+  quaddingCentered,
+  quaddingRightJustified
+};
+
 class Form;
 class FormField;
 class FormFieldButton;
@@ -57,16 +70,11 @@ class FormWidget {
 public:
   virtual ~FormWidget();
 
-  // see the description of FormField::LoadChildrenDefaults
-  virtual void loadDefaults () {}
-  
   // Check if point is inside the field bounding rect
-  GBool inRect(double x, double y)
-    { return x1 <= x && x <= x2 && y1 <= y && y <= y2; }
+  GBool inRect(double x, double y) const;
 
   // Get the field bounding rect
-  void getRect(double *xa1, double *ya1, double *xa2, double *ya2)
-    { *xa1 = x1; *ya1 = y1; *xa2 = x2; *ya2 = y2; }
+  void getRect(double *x1, double *y1, double *x2, double *y2) const;
 
   unsigned getID () { return ID; }
   void setID (unsigned int i) { ID=i; }
@@ -80,30 +88,42 @@ public:
   void setChildNum (unsigned i) { childNum = i; }
   unsigned getChildNum () { return childNum; }
 
-  void setFontSize(double f) { fontSize = f; }
-  double getFontSize () { return fontSize; }
+  double getFontSize() const;
 
-  GBool isModified () { return modified; }
+  GooString *getPartialName() const;
+  GooString *getAlternateUiName() const;
+  GooString *getMappingName() const;
+  GooString *getFullyQualifiedName();
+
+  GBool isModified () const;
 
   bool isReadOnly() const;
 
+  LinkAction *getActivationAction();
+
   // return the unique ID corresponding to pageNum/fieldNum
   static int encodeID (unsigned pageNum, unsigned fieldNum);
   // decode id and retrieve pageNum and fieldNum
   static void decodeID (unsigned id, unsigned* pageNum, unsigned* fieldNum);
 
+  void createWidgetAnnotation();
+  AnnotWidget *getWidgetAnnotation() const { return widget; }
+
+#ifdef DEBUG_FORMS
+  void print(int indent = 0);
+#endif
+
 protected:
-  FormWidget(XRef *xrefA, Object *aobj, unsigned num, Ref aref, FormField *fieldA);
-
-  void updateField (const char *key, Object *value);
+  FormWidget(PDFDoc *docA, Object *aobj, unsigned num, Ref aref, FormField *fieldA);
 
+  AnnotWidget *widget;
   FormField* field;
   FormFieldType type;
   Object obj;
   Ref ref;
+  PDFDoc *doc;
   XRef *xref;
-  GBool defaultsLoaded;
-  GBool modified;
+
   //index of this field in the parent's child list
   unsigned childNum;
 
@@ -116,11 +136,6 @@ protected:
   (decoding) pageNum = id >> 4*sizeof(unsigned); fieldNum = (id << 4*sizeof(unsigned)) >> 4*sizeof(unsigned);
   */
   unsigned ID; 
-
-  double x1, y1;                // lower left corner
-  double x2, y2;                // upper right corner
-  double fontSize; //font size if this widget has text
-
 };
 
 //------------------------------------------------------------------------
@@ -129,17 +144,16 @@ protected:
 
 class FormWidgetButton: public FormWidget {
 public:
-  FormWidgetButton(XRef *xrefA, Object *dict, unsigned num, Ref ref, FormField *p);
+  FormWidgetButton(PDFDoc *docA, Object *dict, unsigned num, Ref ref, FormField *p);
   ~FormWidgetButton ();
 
   FormButtonType getButtonType() const;
   
-  void setState (GBool state, GBool calledByParent=gFalse);
+  void setState (GBool state);
   GBool getState ();
 
-  char* getOnStr () { return onStr->getCString(); }
-
-  void loadDefaults();
+  char* getOnStr();
+  void setAppearanceState(const char *state);
 
   void setNumSiblingsID (int i);
   void setSiblingsID (int i, unsigned id) { siblingsID[i] = id; }
@@ -154,7 +168,6 @@ protected:
   int numSiblingsID;
   GooString *onStr;
   FormFieldButton *parent;
-  GBool state;
 };
 
 //------------------------------------------------------------------------
@@ -163,7 +176,7 @@ protected:
 
 class FormWidgetText: public FormWidget {
 public:
-  FormWidgetText(XRef *xrefA, Object *dict, unsigned num, Ref ref, FormField *p);
+  FormWidgetText(PDFDoc *docA, Object *dict, unsigned num, Ref ref, FormField *p);
   //return the field's content (UTF16BE)
   GooString* getContent() ;
   //return a copy of the field's content (UTF16BE)
@@ -172,8 +185,6 @@ public:
   //except a UTF16BE string
   void setContent(GooString* new_content);
 
-  void loadDefaults ();
-
   bool isMultiline () const; 
   bool isPassword () const; 
   bool isFileSelect () const; 
@@ -192,10 +203,9 @@ protected:
 
 class FormWidgetChoice: public FormWidget {
 public:
-  FormWidgetChoice(XRef *xrefA, Object *dict, unsigned num, Ref ref, FormField *p);
+  FormWidgetChoice(PDFDoc *docA, Object *dict, unsigned num, Ref ref, FormField *p);
   ~FormWidgetChoice();
 
-  void loadDefaults ();
   int getNumChoices();
   //return the display name of the i-th choice (UTF16BE)
   GooString* getChoice(int i);
@@ -223,7 +233,6 @@ public:
   bool commitOnSelChange () const; 
   bool isListBox () const;
 protected:
-  void _updateV ();
   bool _checkRange (int i);
   FormFieldChoice *parent;
 };
@@ -234,7 +243,7 @@ protected:
 
 class FormWidgetSignature: public FormWidget {
 public:
-  FormWidgetSignature(XRef *xrefA, Object *dict, unsigned num, Ref ref, FormField *p);
+  FormWidgetSignature(PDFDoc *docA, Object *dict, unsigned num, Ref ref, FormField *p);
 protected:
   FormFieldSignature *parent;
 };
@@ -248,7 +257,7 @@ protected:
 
 class FormField {
 public:
-  FormField(XRef* xrefa, Object *aobj, const Ref& aref, FormFieldType t=formUndef);
+  FormField(PDFDoc *docA, Object *aobj, const Ref& aref, FormField *parent, std::set<int> *usedParents, FormFieldType t=formUndef);
 
   virtual ~FormField();
 
@@ -260,27 +269,57 @@ public:
   void setReadOnly (bool b) { readOnly = b; }
   bool isReadOnly () const { return readOnly; }
 
+  GBool isModified () const;
+
+  GooString* getDefaultAppearance() const { return defaultAppearance; }
+  GBool hasTextQuadding() const { return hasQuadding; }
+  VariableTextQuadding getTextQuadding() const { return quadding; }
+
+  GooString *getPartialName() const { return partialName; }
+  GooString *getAlternateUiName() const { return alternateUiName; }
+  GooString *getMappingName() const { return mappingName; }
+  GooString *getFullyQualifiedName();
+
   FormWidget* findWidgetByRef (Ref aref);
-  // Since while loading their defaults, children may call parents methods, it's better
-  // to do that when parents are completly constructed
-  void loadChildrenDefaults();
+  FormWidget *getWidget(int i) { return terminal ? widgets[i] : NULL; }
 
   // only implemented in FormFieldButton
   virtual void fillChildrenSiblingsID ();
 
+  void createWidgetAnnotations();
+
+#ifdef DEBUG_FORMS
+  void printTree(int indent = 0);
+  virtual void print(int indent = 0);
+#endif
+
 
  protected:
   void _createWidget (Object *obj, Ref aref);
+  void createChildren(std::set<int> *usedParents);
 
   FormFieldType type;           // field type
   Ref ref;
   bool terminal;
   Object obj;
+  PDFDoc *doc;
   XRef *xref;
   FormField **children;
+  FormField *parent;
   int numChildren;
   FormWidget **widgets;
   bool readOnly;
+  GBool modified;
+
+  GooString *partialName; // T field
+  GooString *alternateUiName; // TU field
+  GooString *mappingName; // TM field
+  GooString *fullyQualifiedName;
+
+  // Variable Text
+  GooString *defaultAppearance;
+  GBool hasQuadding;
+  VariableTextQuadding quadding;
 
 private:
   FormField() {}
@@ -293,23 +332,33 @@ private:
 
 class FormFieldButton: public FormField {
 public:
-  FormFieldButton(XRef *xrefA, Object *dict, const Ref& ref);
+  FormFieldButton(PDFDoc *docA, Object *dict, const Ref& ref, FormField *parent, std::set<int> *usedParents);
 
   FormButtonType getButtonType () { return btype; }
 
   bool noToggleToOff () const { return noAllOff; }
 
   // returns gTrue if the state modification is accepted
-  GBool setState (int num, GBool s);
-  
+  GBool setState (char *state);
+  GBool getState(char *state);
+
+  char *getAppearanceState() { return appearanceState.isName() ? appearanceState.getName() : NULL; }
+
   void fillChildrenSiblingsID ();
 
+#ifdef DEBUG_FORMS
+  void print(int indent = 0);
+#endif
+
   virtual ~FormFieldButton();
 protected:
+  void updateState(char *state);
+
   FormButtonType btype;
   int size;
   int active_child; //only used for combo box
   bool noAllOff;
+  Object appearanceState; // V
 };
 
 //------------------------------------------------------------------------
@@ -318,7 +367,7 @@ protected:
 
 class FormFieldText: public FormField {
 public:
-  FormFieldText(XRef *xrefA, Object *dict, const Ref& ref);
+  FormFieldText(PDFDoc *docA, Object *dict, const Ref& ref, FormField *parent, std::set<int> *usedParents);
   
   GooString* getContent () { return content; }
   GooString* getContentCopy ();
@@ -334,6 +383,10 @@ public:
   bool isRichText () const { return richText; }
 
   int getMaxLen () const { return maxLen; }
+
+#ifdef DEBUG_FORMS
+  void print(int indent = 0);
+#endif
 protected:
   GooString* content;
   bool multiline;
@@ -352,13 +405,15 @@ protected:
 
 class FormFieldChoice: public FormField {
 public:
-  FormFieldChoice(XRef *xrefA, Object *aobj, const Ref& ref);
+  FormFieldChoice(PDFDoc *docA, Object *aobj, const Ref& ref, FormField *parent, std::set<int> *usedParents);
 
   virtual ~FormFieldChoice();
 
   int getNumChoices() { return numChoices; }
-  GooString* getChoice(int i) { return choices[i].optionName; }
-  GooString* getExportVal (int i) { return choices[i].exportVal; }
+  GooString* getChoice(int i) { return choices ? choices[i].optionName : NULL; }
+  GooString* getExportVal (int i) { return choices ? choices[i].exportVal : NULL; }
+  // For multi-select choices it returns the first one
+  GooString* getSelectedChoice();
 
   //select the i-th choice
   void select (int i); 
@@ -385,13 +440,16 @@ public:
   bool commitOnSelChange () const { return doCommitOnSelChange; }
   bool isListBox () const { return !combo; }
 
-  /* these functions _must_ only be used by FormWidgetChoice */
-  void _setNumChoices (int i) { numChoices = i; }
-  void _createChoicesTab ();
-  void _setChoiceExportVal (int i, GooString* str) { choices[i].exportVal = str; }
-  void _setChoiceOptionName (int i, GooString* str) { choices[i].optionName = str; }
+  int getTopIndex() const { return topIdx; }
+
+#ifdef DEBUG_FORMS
+  void print(int indent = 0);
+#endif
 
 protected:
+  void unselectAll();
+  void updateSelection();
+
   bool combo;
   bool edit;
   bool multiselect;
@@ -406,7 +464,8 @@ protected:
 
   int numChoices;
   ChoiceOpt* choices;
-  GooString* editedChoice; 
+  GooString* editedChoice;
+  int topIdx; // TI
 };
 
 //------------------------------------------------------------------------
@@ -415,9 +474,13 @@ protected:
 
 class FormFieldSignature: public FormField {
 public:
-  FormFieldSignature(XRef *xrefA, Object *dict, const Ref& ref);
+  FormFieldSignature(PDFDoc *docA, Object *dict, const Ref& ref, FormField *parent, std::set<int> *usedParents);
 
   virtual ~FormFieldSignature();
+
+#ifdef DEBUG_FORMS
+  void print(int indent = 0);
+#endif
 };
 
 //------------------------------------------------------------------------
@@ -428,21 +491,25 @@ public:
 
 class Form {
 public:
-  Form(XRef *xrefA, Object* acroForm);
+  Form(PDFDoc *docA, Object* acroForm);
 
   ~Form();
 
   // Look up an inheritable field dictionary entry.
-  static Object *fieldLookup(Dict *field, char *key, Object *obj);
+  static Object *fieldLookup(Dict *field, const char *key, Object *obj);
   
   /* Creates a new Field of the type specified in obj's dict.
      used in Form::Form and FormField::FormField */
-  static FormField *createFieldFromDict (Object* obj, XRef *xref, const Ref& aref);
+  static FormField *createFieldFromDict (Object* obj, PDFDoc *docA, const Ref& aref, FormField *parent, std::set<int> *usedParents);
 
   Object *getObj () const { return acroForm; }
   GBool getNeedAppearances () const { return needAppearances; }
   int getNumFields() const { return numFields; }
   FormField* getRootField(int i) const { return rootFields[i]; }
+  GooString* getDefaultAppearance() const { return defaultAppearance; }
+  VariableTextQuadding getTextQuadding() const { return quadding; }
+  GfxResources* getDefaultResources() const { return defaultResources; }
+  Object* getDefaultResourcesObj() { return &resDict; }
 
   FormWidget* findWidgetByRef (Ref aref);
 
@@ -451,9 +518,16 @@ private:
   FormField** rootFields;
   int numFields;
   int size;
+  PDFDoc *doc;
   XRef* xref;
   Object *acroForm;
   GBool needAppearances;
+  GfxResources *defaultResources;
+  Object resDict;
+
+  // Variable Text
+  GooString *defaultAppearance;
+  VariableTextQuadding quadding;
 };
 
 //------------------------------------------------------------------------
@@ -462,7 +536,7 @@ private:
 
 class FormPageWidgets {
 public:
-  FormPageWidgets (XRef *xrefA, Object* annots, unsigned int page, Form *form);
+  FormPageWidgets (Annots* annots, unsigned int page, Form *form);
   ~FormPageWidgets();
   
   int getNumWidgets() const { return numWidgets; }
@@ -472,8 +546,6 @@ private:
   FormWidget** widgets;
   int numWidgets;
   int size;
-  unsigned pageNum;
-  XRef* xref;
 };
 
 #endif
