@@ -4,6 +4,8 @@
 //
 // Copyright (C) 2005-2006 Kristian Høgsberg <krh@redhat.com>
 // Copyright (C) 2005, 2009 Albert Astals Cid <aacid@kde.org>
+// Copyright (C) 2011 Simon Kellner <kellner@kit.edu>
+// Copyright (C) 2012 Fabio D'Urso <fabiodurso@hotmail.it>
 //
 // To see a description of the changes please see the Changelog file that
 // came with your tarball or type make ChangeLog if you are building from git
@@ -258,11 +260,11 @@ GBool PageLabelInfo::labelToIndex(GooString *label, int *index)
   Interval *interval;
   char *str = label->getCString(), *end;
   int prefixLength;
-  int i, base, number;
+  int i, number;
 
-  base = 0;
   for (i = 0; i < intervals.getLength(); i++) {
     interval = (Interval *) intervals.get(i);
+    const int base = interval->base;
     prefixLength = interval->prefix->getLength();
     if (label->cmpN(interval->prefix, prefixLength) != 0)
       continue;
@@ -294,8 +296,6 @@ GBool PageLabelInfo::labelToIndex(GooString *label, int *index)
     case Interval::None:
       break;
     }
-
-    base += interval->length;
   }
 
   return gFalse;
@@ -353,8 +353,6 @@ GBool PageLabelInfo::indexToLabel(int index, GooString *label)
 	  ucs2_char[1] = number_string.getChar(i);
 	  label->append(ucs2_char, 2);
       }
-      ucs2_char[1] = 0;
-      label->append(ucs2_char, 2);
   } else {
       label->append(&number_string);
   }
