@@ -6,6 +6,20 @@
 //
 //========================================================================
 
+//========================================================================
+//
+// Modified under the Poppler project - http://poppler.freedesktop.org
+//
+// All changes made under the Poppler project to this file are licensed
+// under GPL version 2 or later
+//
+// Copyright (C) 2010, 2012 Albert Astals Cid <aacid@kde.org>
+//
+// To see a description of the changes please see the Changelog file that
+// came with your tarball or type make ChangeLog if you are building from git
+//
+//========================================================================
+
 #include <config.h>
 
 #ifdef USE_GCC_PRAGMAS
@@ -17,16 +31,13 @@
 #include "Decrypt.h"
 #include "Error.h"
 #include "GlobalParams.h"
-#if HAVE_XPDFCORE
-#  include "XPDFCore.h"
-#elif HAVE_WINPDFCORE
-#  include "WinPDFCore.h"
-#endif
 #ifdef ENABLE_PLUGINS
 #  include "XpdfPluginAPI.h"
 #endif
 #include "SecurityHandler.h"
 
+#include <limits.h>
+
 //------------------------------------------------------------------------
 // SecurityHandler
 //------------------------------------------------------------------------
@@ -47,14 +58,15 @@ SecurityHandler *SecurityHandler::make(PDFDoc *docA, Object *encryptDictA) {
       secHdlr = new ExternalSecurityHandler(docA, encryptDictA, xsh);
     } else {
 #endif
-      error(-1, "Couldn't find the '%s' security handler",
+      error(errSyntaxError, -1, "Couldn't find the '{0:s}' security handler",
 	    filterObj.getName());
       secHdlr = NULL;
 #ifdef ENABLE_PLUGINS
     }
 #endif
   } else {
-    error(-1, "Missing or invalid 'Filter' entry in encryption dictionary");
+    error(errSyntaxError, -1,
+	  "Missing or invalid 'Filter' entry in encryption dictionary");
     secHdlr = NULL;
   }
   filterObj.free();
@@ -93,7 +105,7 @@ GBool SecurityHandler::checkEncryption(GooString *ownerPassword,
     }
   }
   if (!ok) {
-    error(-1, "Incorrect password");
+    error(errCommandLine, -1, "Incorrect password");
   }
   return ok;
 }
@@ -128,8 +140,8 @@ StandardSecurityHandler::StandardSecurityHandler(PDFDoc *docA,
   SecurityHandler(docA)
 {
   Object versionObj, revisionObj, lengthObj;
-  Object ownerKeyObj, userKeyObj, permObj, fileIDObj;
-  Object fileIDObj1;
+  Object ownerKeyObj, userKeyObj, ownerEncObj, userEncObj;
+  Object permObj, fileIDObj, fileIDObj1;
   Object cryptFiltersObj, streamFilterObj, stringFilterObj;
   Object cryptFilterObj, cfmObj, cfLengthObj;
   Object encryptMetadataObj;
@@ -138,101 +150,158 @@ StandardSecurityHandler::StandardSecurityHandler(PDFDoc *docA,
   fileID = NULL;
   ownerKey = NULL;
   userKey = NULL;
+  ownerEnc = NULL;
+  userEnc = NULL;
+  fileKeyLength = 0;
 
   encryptDictA->dictLookup("V", &versionObj);
   encryptDictA->dictLookup("R", &revisionObj);
   encryptDictA->dictLookup("Length", &lengthObj);
   encryptDictA->dictLookup("O", &ownerKeyObj);
   encryptDictA->dictLookup("U", &userKeyObj);
+  encryptDictA->dictLookup("OE", &ownerEncObj);
+  encryptDictA->dictLookup("UE", &userEncObj);
   encryptDictA->dictLookup("P", &permObj);
+  if (permObj.isUint()) {
+      unsigned int permUint = permObj.getUint();
+      int perms = permUint - UINT_MAX - 1;
+      permObj.free();
+      permObj.initInt(perms);
+  }
   doc->getXRef()->getTrailerDict()->dictLookup("ID", &fileIDObj);
   if (versionObj.isInt() &&
       revisionObj.isInt() &&
-      ownerKeyObj.isString() && ownerKeyObj.getString()->getLength() == 32 &&
-      userKeyObj.isString() && userKeyObj.getString()->getLength() == 32 &&
-      permObj.isInt()) {
+      permObj.isInt() &&
+      ownerKeyObj.isString() &&
+      userKeyObj.isString()) {
     encVersion = versionObj.getInt();
     encRevision = revisionObj.getInt();
-    encAlgorithm = cryptRC4;
-    // revision 2 forces a 40-bit key - some buggy PDF generators
-    // set the Length value incorrectly
-    if (encRevision == 2 || !lengthObj.isInt()) {
-      fileKeyLength = 5;
-    } else {
-      fileKeyLength = lengthObj.getInt() / 8;
-    }
-    encryptMetadata = gTrue;
-    //~ this currently only handles a subset of crypt filter functionality
-    if (encVersion == 4 && encRevision == 4) {
-      encryptDictA->dictLookup("CF", &cryptFiltersObj);
-      encryptDictA->dictLookup("StmF", &streamFilterObj);
-      encryptDictA->dictLookup("StrF", &stringFilterObj);
-      if (cryptFiltersObj.isDict() &&
-	  streamFilterObj.isName() &&
-	  stringFilterObj.isName() &&
-	  !strcmp(streamFilterObj.getName(), stringFilterObj.getName())) {
-	if (cryptFiltersObj.dictLookup(streamFilterObj.getName(),
-				       &cryptFilterObj)->isDict()) {
-	  cryptFilterObj.dictLookup("CFM", &cfmObj);
-	  if (cfmObj.isName("V2")) {
-	    encVersion = 2;
-	    encRevision = 3;
-	    if (cryptFilterObj.dictLookup("Length", &cfLengthObj)->isInt()) {
-	      //~ according to the spec, this should be cfLengthObj / 8
-	      fileKeyLength = cfLengthObj.getInt();
+    if ((encRevision <= 4 &&
+	 ownerKeyObj.getString()->getLength() == 32 &&
+	 userKeyObj.getString()->getLength() == 32) ||
+	(encRevision == 5 &&
+	 // the spec says 48 bytes, but Acrobat pads them out longer
+	 ownerKeyObj.getString()->getLength() >= 48 &&
+	 userKeyObj.getString()->getLength() >= 48 &&
+	 ownerEncObj.isString() &&
+	 ownerEncObj.getString()->getLength() == 32 &&
+	 userEncObj.isString() &&
+	 userEncObj.getString()->getLength() == 32)) {
+      encAlgorithm = cryptRC4;
+      // revision 2 forces a 40-bit key - some buggy PDF generators
+      // set the Length value incorrectly
+      if (encRevision == 2 || !lengthObj.isInt()) {
+	fileKeyLength = 5;
+      } else {
+	fileKeyLength = lengthObj.getInt() / 8;
+      }
+      encryptMetadata = gTrue;
+      //~ this currently only handles a subset of crypt filter functionality
+      //~ (in particular, it ignores the EFF entry in encryptDictA, and
+      //~ doesn't handle the case where StmF, StrF, and EFF are not all the
+      //~ same)
+      if ((encVersion == 4 || encVersion == 5) &&
+	  (encRevision == 4 || encRevision == 5)) {
+	encryptDictA->dictLookup("CF", &cryptFiltersObj);
+	encryptDictA->dictLookup("StmF", &streamFilterObj);
+	encryptDictA->dictLookup("StrF", &stringFilterObj);
+	if (cryptFiltersObj.isDict() &&
+	    streamFilterObj.isName() &&
+	    stringFilterObj.isName() &&
+	    !strcmp(streamFilterObj.getName(), stringFilterObj.getName())) {
+	  if (!strcmp(streamFilterObj.getName(), "Identity")) {
+	    // no encryption on streams or strings
+	    encVersion = encRevision = -1;
+	  } else {
+	    if (cryptFiltersObj.dictLookup(streamFilterObj.getName(),
+					   &cryptFilterObj)->isDict()) {
+	      cryptFilterObj.dictLookup("CFM", &cfmObj);
+	      if (cfmObj.isName("V2")) {
+		encVersion = 2;
+		encRevision = 3;
+		if (cryptFilterObj.dictLookup("Length",
+					      &cfLengthObj)->isInt()) {
+		  //~ according to the spec, this should be cfLengthObj / 8
+		  fileKeyLength = cfLengthObj.getInt();
+		}
+		cfLengthObj.free();
+	      } else if (cfmObj.isName("AESV2")) {
+		encVersion = 2;
+		encRevision = 3;
+		encAlgorithm = cryptAES;
+		if (cryptFilterObj.dictLookup("Length",
+					      &cfLengthObj)->isInt()) {
+		  //~ according to the spec, this should be cfLengthObj / 8
+		  fileKeyLength = cfLengthObj.getInt();
+		}
+		cfLengthObj.free();
+	      } else if (cfmObj.isName("AESV3")) {
+		encVersion = 5;
+		encRevision = 5;
+		encAlgorithm = cryptAES256;
+		if (cryptFilterObj.dictLookup("Length",
+					      &cfLengthObj)->isInt()) {
+		  //~ according to the spec, this should be cfLengthObj / 8
+		  fileKeyLength = cfLengthObj.getInt();
+		}
+		cfLengthObj.free();
+	      }
+	      cfmObj.free();
 	    }
-	    cfLengthObj.free();
-	  } else if (cfmObj.isName("AESV2")) {
-	    encVersion = 2;
-	    encRevision = 3;
-	    encAlgorithm = cryptAES;
-	    if (cryptFilterObj.dictLookup("Length", &cfLengthObj)->isInt()) {
-	      //~ according to the spec, this should be cfLengthObj / 8
-	      fileKeyLength = cfLengthObj.getInt();
-	    }
-	    cfLengthObj.free();
+	    cryptFilterObj.free();
 	  }
-	  cfmObj.free();
 	}
-	cryptFilterObj.free();
+	stringFilterObj.free();
+	streamFilterObj.free();
+	cryptFiltersObj.free();
+	if (encryptDictA->dictLookup("EncryptMetadata",
+				     &encryptMetadataObj)->isBool()) {
+	  encryptMetadata = encryptMetadataObj.getBool();
+	}
+	encryptMetadataObj.free();
       }
-      stringFilterObj.free();
-      streamFilterObj.free();
-      cryptFiltersObj.free();
-      if (encryptDictA->dictLookup("EncryptMetadata",
-				   &encryptMetadataObj)->isBool()) {
-	encryptMetadata = encryptMetadataObj.getBool();
-      }
-      encryptMetadataObj.free();
-    }
-    permFlags = permObj.getInt();
-    ownerKey = ownerKeyObj.getString()->copy();
-    userKey = userKeyObj.getString()->copy();
-    if (encVersion >= 1 && encVersion <= 2 &&
-	encRevision >= 2 && encRevision <= 3) {
-      if (fileIDObj.isArray()) {
-	if (fileIDObj.arrayGet(0, &fileIDObj1)->isString()) {
-	  fileID = fileIDObj1.getString()->copy();
+      permFlags = permObj.getInt();
+      ownerKey = ownerKeyObj.getString()->copy();
+      userKey = userKeyObj.getString()->copy();
+      if (encVersion >= 1 && encVersion <= 2 &&
+	  encRevision >= 2 && encRevision <= 3) {
+	if (fileIDObj.isArray()) {
+	  if (fileIDObj.arrayGet(0, &fileIDObj1)->isString()) {
+	    fileID = fileIDObj1.getString()->copy();
+	  } else {
+	    fileID = new GooString();
+	  }
+	  fileIDObj1.free();
 	} else {
 	  fileID = new GooString();
 	}
-	fileIDObj1.free();
-      } else {
-	fileID = new GooString();
+	if (fileKeyLength > 16 || fileKeyLength < 0) {
+	  fileKeyLength = 16;
+	}
+	ok = gTrue;
+      } else if (encVersion == 5 && encRevision == 5) {
+	fileID = new GooString(); // unused for V=R=5
+	ownerEnc = ownerEncObj.getString()->copy();
+	userEnc = userEncObj.getString()->copy();
+	if (fileKeyLength > 32 || fileKeyLength < 0) {
+	  fileKeyLength = 32;
+	}
+	ok = gTrue;
+      } else if (!(encVersion == -1 && encRevision == -1)) {
+	error(errUnimplemented, -1,
+	      "Unsupported version/revision (%d/%d) of Standard security handler",
+	      encVersion, encRevision);
       }
-      ok = gTrue;
     } else {
-      error(-1, "Unsupported version/revision (%d/%d) of Standard security handler",
-	    encVersion, encRevision);
+      error(errSyntaxError, -1, "Invalid encryption key length");
     }
   } else {
-    error(-1, "Weird encryption info");
-  }
-  if (fileKeyLength > 16) {
-    fileKeyLength = 16;
+    error(errSyntaxError, -1, "Weird encryption info");
   }
   fileIDObj.free();
   permObj.free();
+  userEncObj.free();
+  ownerEncObj.free();
   userKeyObj.free();
   ownerKeyObj.free();
   lengthObj.free();
@@ -250,6 +319,16 @@ StandardSecurityHandler::~StandardSecurityHandler() {
   if (userKey) {
     delete userKey;
   }
+  if (ownerEnc) {
+    delete ownerEnc;
+  }
+  if (userEnc) {
+    delete userEnc;
+  }
+}
+
+GBool StandardSecurityHandler::isUnencrypted() {
+  return encVersion == -1 && encRevision == -1;
 }
 
 void *StandardSecurityHandler::makeAuthData(GooString *ownerPassword,
@@ -261,27 +340,7 @@ void *StandardSecurityHandler::makeAuthData(GooString *ownerPassword,
 }
 
 void *StandardSecurityHandler::getAuthData() {
-#if HAVE_XPDFCORE
-  XPDFCore *core;
-  GooString *password;
-
-  if (!(core = (XPDFCore *)doc->getGUIData()) ||
-      !(password = core->getPassword())) {
-    return NULL;
-  }
-  return new StandardAuthData(password, password->copy());
-#elif HAVE_WINPDFCORE
-  WinPDFCore *core;
-  GooString *password;
-
-  if (!(core = (WinPDFCore *)doc->getGUIData()) ||
-      !(password = core->getPassword())) {
-    return NULL;
-  }
-  return new StandardAuthData(password, password->copy());
-#else
   return NULL;
-#endif
 }
 
 void StandardSecurityHandler::freeAuthData(void *authData) {
@@ -302,7 +361,8 @@ GBool StandardSecurityHandler::authorize(void *authData) {
     userPassword = NULL;
   }
   if (!Decrypt::makeFileKey(encVersion, encRevision, fileKeyLength,
-			    ownerKey, userKey, permFlags, fileID,
+			    ownerKey, userKey, ownerEnc, userEnc,
+			    permFlags, fileID,
 			    ownerPassword, userPassword, fileKey,
 			    encryptMetadata, &ownerPasswordOk)) {
     return gFalse;
@@ -376,7 +436,7 @@ GBool ExternalSecurityHandler::authorize(void *authData) {
   if (!(permFlags & xpdfPermissionOpen)) {
     return gFalse;
   }
-  if (!(*xsh->getKey)(xsh->handlerData, docData, &key, &length, &encVersion)) {
+  if (!(*xsh->getKey)(xsh->handlerData, docData, &key, &length, &encVersion, &encRevision)) {
     return gFalse;
   }
   if ((fileKeyLength = length) > 16) {
