@@ -11,9 +11,11 @@
 // All changes made under the Poppler project to this file are licensed
 // under GPL version 2 or later
 //
-// Copyright (C) 2006 Albert Astals Cid <aacid@kde.org>
+// Copyright (C) 2006, 2010 Albert Astals Cid <aacid@kde.org>
 // Copyright (C) 2008 Tomas Are Haavet <tomasare@gmail.com>
-// Copyright (C) 2009 Thomas Freitag <Thomas.Freitag@alfa.de>
+// Copyright (C) 2009, 2011 Thomas Freitag <Thomas.Freitag@alfa.de>
+// Copyright (C) 2009 Stefan Thomas <thomas@eload24.com>
+// Copyright (C) 2010 William Bader <williambader@hotmail.com>
 //
 // To see a description of the changes please see the Changelog file that
 // came with your tarball or type make ChangeLog if you are building from git
@@ -32,6 +34,8 @@
 #if USE_FIXEDPOINT
 #include "goo/FixedPoint.h"
 typedef FixedPoint SplashCoord;
+#elif USE_FLOAT
+typedef float SplashCoord;
 #else
 typedef double SplashCoord;
 #endif
@@ -150,4 +154,16 @@ struct SplashScreenParams {
 
 typedef int SplashError;
 
+
+//------------------------------------------------------------------------
+// image file formats
+//------------------------------------------------------------------------
+
+enum SplashImageFileFormat {
+  splashFormatJpeg,
+  splashFormatPng,
+  splashFormatTiff,
+  splashFormatJpegCMYK
+};
+
 #endif
